class BossCharger {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //moving
  //collection of when the monster should charge toward the player
  int chargeTime=30;  //how long it should take to reach the player in frames
  int retreatTime=30;
  float chargeX=180;  //once the boss hits this x value, the charge is done

  //returning from a charge
  float waitYPadding=100;  //how far from the top and bottom it can be
  float waitX=width-180;  //goal x Point when returning from a charge
  int waitTime=50;        //NOT USED aproximate number of frames it should be waiting there 

  //these are set when a charge starts or ends
  float xVel=0;
  float yVel=0;  

  //getting hit
  float hitX=-70;    //offset from regular x
  float hitY=30;   //offset from regular y
  float hitW=140;
  float hitH=50;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showing the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=2;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=50;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  float damage=3;  //damage form touching the boss

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int moveTimes[];
  int nextMoveTime=0;

  BossCharger() {
    x=waitX;
    y=height/2; 

    pic=loadImage("data/pics/chargeMonster.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      color col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    setTimes();
  }

  void update(int time) {

    if (!gameOver)
      checkTiming(time);

    //if the beats is over the line, start retreating
    if (x<chargeX) {
      //claculate a random y point to try and return to
      float targetY=random(waitYPadding, height-waitYPadding);

      //set the velocities to try to hit that point
      xVel=(waitX-x)/retreatTime;
      yVel=(targetY-y)/retreatTime;
    }

    //if it has crossed the wait line, just wait
    if (x>waitX) {
      x=waitX;
      xVel=0;
      yVel=0;
    }

    x+=xVel;
    y+=yVel;


    //if this is at all visible, it is active
    active=raisingY>0;

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;

      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //adjust the mask for the image
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1, millis()*0.001)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //lower the hitTimer
    hitTimer--;
  }

  void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //see if it is time to charge
    if (nextMoveTime<moveTimes.length) {
      if (time>moveTimes[nextMoveTime]) {
        //charge at player 
        //calculate the speed to move to reahc player pos in the given amount of time
        xVel=(ship.x-x)/chargeTime;
        yVel=(ship.y-y)/chargeTime;

        nextMoveTime++;  //increase to the next time in the arra
      }
    }
  }

  void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

  }

  //called when the boss gets hurt
  void hit() {
    hitTimer=hitTime;
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }


  //causes the monster to sink or rise
  void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }
  
   //hard code the times that this thing needs to charge
  void setTimes() {
    raiseTimes=new int[2];
    lowerTimes=new int[2];

    raiseTimes[0]=16050;
    lowerTimes[0]=25500;
    raiseTimes[1]=81500;
    lowerTimes[1]=107600;
    
    //use an arryalist for now because it is easier to ad to
    ArrayList<Integer> moveTimesList=new ArrayList<Integer>();
    
    moveTimesList.add(82570);
    moveTimesList.add(85720);
    moveTimesList.add(88910);
    moveTimesList.add(92020);
    moveTimesList.add(95200);
    moveTimesList.add(98160);
    
    moveTimesList.add(101330);
    moveTimesList.add(102100);
    moveTimesList.add(102870);
    moveTimesList.add(103680);
    
    moveTimesList.add(104450);
    moveTimesList.add(105260);
    
    moveTimesList.add(99999999);  //put the last charge at a time that can never happen
    //convert the arraylist to good old array
    moveTimes=new int[moveTimesList.size()];
    for (int i=0; i<moveTimesList.size(); i++){
      moveTimes[i]= moveTimesList.get(i).intValue();
    }
    
  }
  
  //advances all timers to this point in the song
  void skipTo(int time) {
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;

    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;

    for (int i=0; i<moveTimes.length;i++)
      if (time > moveTimes[i])  nextMoveTime=i;
  }
  
}

