class BossShooter {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //how much to slow down the values coming into the noise function
  float noiseSpeedX=0.00018; 
  float noiseSpeedY=0.0002;  
  //where in the board it will be
  float xRange=200;
  float yRange=300;
  float xCenter=width-400;
  float yCenter=height/2;
  //keep it moving up and down
  //y locaiton from noise will be centerred around this
  float ySpeed=0.0002;
  float yPadding=100;  //keep the monster on screen

  //shooting
  float yShotOffset=-30;
  float xShotOffset=-30;
  boolean shootHigh=true;  //boss should alternate shooting high and low
  boolean shooting=false;  //flag for main to shoot a bullet

  //getting hit
  float hitX=25;    //offset from regular x
  float hitY=10;   //offset from regular y
  float hitW=40;
  float hitH=100;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showin the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=4;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=10;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  float damage=3;  //damage from touching the beast

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int shotTimes[];
  int nextShotTime=0;

  BossShooter() {
    x=xCenter;
    y=yCenter; 

    pic=loadImage("data/pics/shooterBoss.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      color col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    setTiming();
  }

  void update(int time) {
    //keep it moving up and down
    yCenter=height/2+sin(millis()*ySpeed)*yPadding;

    x=xCenter+noise((float)millis()*noiseSpeedX, 1)*xRange*2-xRange;
    y=yCenter+noise((float)millis()*noiseSpeedY, 1)*yRange*2-yRange;
    //the 1 just serves to seperate if from the noise being generated for the other monsters

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;
      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //if this is at all visible, it is active
    active=raisingY>0;

    //adjust the mask for the beast
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1, millis()*0.001)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //lower the hitTimer
    hitTimer--;

    if (!gameOver)
      checkTiming(time);  //see if anything needs to happen based on timing
  }

  void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //time to shoot?
    if (nextShotTime<shotTimes.length) {
      if (time>shotTimes[nextShotTime]) {
        shoot();
        nextShotTime++;
      }
    }
  }

  void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

  }

  //does the animaiton for shooting
  //firing the bullet is handled in main
  void shoot() { 
    shootHigh=!shootHigh;
    shooting=true;
  }

  //called when the boss gets hurt
  void hit() {
    hitTimer=hitTime;
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }

  //causes the monster to sink or rise
  void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }

  //hardcoding all of the timing info for the boss in millis
  void setTiming() {
    raiseTimes=new int[3];
    lowerTimes=new int[3];

    raiseTimes[0] =12900;
    lowerTimes[0] =25500;
    raiseTimes[1] =31400;
    lowerTimes[1] =69500;
    raiseTimes[2]=138800;
    lowerTimes[2]=162800;

    //fill up the shooting array list then convert to normal array
    ArrayList<Integer> shotTimesList=new ArrayList<Integer>();
    
    //first segment starts at 31.84 and occurs every 0.4 seconds
    //THIS PART WAS CODED DIFFERENTLY THAN THE OTHERS BECAUSE I DID IT FIRST
    int time=69500-31840;
    int shotSpacing=400;
    int numShots=time/shotSpacing;
    for (int i=0; i<numShots; i++)
      shotTimesList.add(31840+i*shotSpacing);
      
      
    //second segment
    int startTime=139200;
    int endTime=  150250;
    shotSpacing=400;
    
    for (int i=startTime; i<endTime; i+=shotSpacing)
      shotTimesList.add(i);
      
    //quick series of guitar hits then a  pause
    startTime=endTime;
    endTime=151130;
    shotSpacing=100;
    for (int i=startTime; i<endTime; i+=shotSpacing)
      shotTimesList.add(i);
      
      
    //a few series of 5 notes
    //timing within the cycle
    startTime=151900;  //115650;
    endTime=159000;  //when the guitars end
    int cycleTime=3181;  //how long it takes to go through a full cycle
    
    int cycleTiming[]=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=1331;
    cycleTiming[2]=1493;
    cycleTiming[3]=2190;
    cycleTiming[4]=2337;
    
    for (int i=startTime; i<endTime; i+=cycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<5; c++) {
        shotTimesList.add(i+cycleTiming[c]);
      }
    }
    
    //2 more to finish it out
    shotTimesList.add(161360);
    shotTimesList.add(161460);
    
      
      
    //convert the arraylist to good old array
    shotTimes=new int[shotTimesList.size()];
    for (int i=0; i<shotTimesList.size(); i++){
      shotTimes[i]= shotTimesList.get(i).intValue();
    }
  }
  
  //advances all timers to this point in the song
  void skipTo(int time){
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;
    
    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;
      
    for (int i=0; i<shotTimes.length; i++)
      if (time > shotTimes[i])  nextShotTime=i;
  }
  
}

