class BossWhirlPool {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //how much to slow down the values coming into the noise function
  float noiseSpeedX=0.0003; 
  float noiseSpeedY=0.0002;  
  //where in the board it will be
  float xRange=300;
  float yRange=300;
  float xCenter=width-400;
  float yCenter=height/2;
  //keep it moving up and down
  //y locaiton from noise will be centerred around this
  float ySpeed=0.0002;
  float yPadding=100;  //keep the monster on screen
  //x location from noise will be centerred aorund this 
  float xSpeed=0.0007;
  float xPadding=100;  //keep the monster on screen
  float xShift=-150;  //how much to offset the track

  //getting hit
  float hitX=15;    //offset from regular x
  float hitY=10;   //offset from regular y
  float hitW=60;
  float hitH=140;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showing the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=4;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=10;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  //whirlpools
  WhirlPool whirlpools[]=new WhirlPool[6];
  float whirlpoolDamage=2;
  float damage=3;  //damage from touching the beast

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int whirlTimes[];
  int whirlStrength[];
  int nextWhirlTime=0;

  BossWhirlPool() {
    x=xCenter;
    y=yCenter; 

    pic=loadImage("data/pics/whirlpoolMonster.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      color col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    //set up the whirlPools
    float angleSpacing=TWO_PI/whirlpools.length;
    for (int i=0; i<whirlpools.length; i++) {
      whirlpools[i]=new WhirlPool(i*angleSpacing);
    }

    //set timing
    setTiming();
  }

  void update(int time) {
    //keep it moving up and down
    yCenter=height/2+sin(millis()*ySpeed)*yPadding;
    xCenter=width/2+sin(millis()*xSpeed)*xPadding +xShift;
    //move it around that center point
    x=xCenter+noise((float)millis()*noiseSpeedX, 1000)*xRange*2-xRange;
    y=yCenter+noise((float)millis()*noiseSpeedY, 1000)*yRange*2-yRange;
    //the 1000 just serves to seperate if from the noise being generated for the other monsters

    //if this is at all visible, it is active
    active=raisingY>0;

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;

      //set the whirlpool size
      for (int i=0; i<whirlpools.length; i++) {
        whirlpools[i].imageScale=map(raisingY, 0, pic.height-raisingPadding, 0, 1);
      }

      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //adjust the mask for the image
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1, millis()*0.001)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //check for timing related events
    if (!gameOver)
      checkTiming(time);

    //lower the hitTimer
    hitTimer--;

    //update the whirlpools
    for (int i=0; i<whirlpools.length; i++)
      whirlpools[i].update(x, y);
  }

  void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //time to update the whirlpools?
    if (nextWhirlTime<whirlTimes.length) {
      if (time>whirlTimes[nextWhirlTime]) {
        //set all fo the whilr pools
        setPools(whirlStrength[nextWhirlTime]);
        nextWhirlTime++;
      }
    }
  }

  void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

    //draw the whirlpools
    if (active) {
      tint(255, 255, 255);
      for (int i=0; i<whirlpools.length; i++)
        whirlpools[i].draw();
    }

  }

  //called when the boss gets hurt
  void hit() {
    hitTimer=hitTime;
  }

  //sets the distance on the pools
  //accepts a number between 0 and 1
  void setPools(float poolDist) {
    for (int i=0; i<whirlpools.length; i++)
      whirlpools[i].setDist(poolDist);
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }

  //checks to see if the ship hit a whirlpool
  //takes the ship's hit box in the form of the upper left and lower right corners
  boolean whirlPoolHitCheck(float x1, float x2, float y1, float y2) {
    //if the monster is not out of the water yet, don't let the whirlpools hurt the player
    if (raisingY<pic.height-raisingPadding)  return false;

    for (int i=0; i<whirlpools.length; i++) {
      if (whirlpools[i].checkHit(x1, x2, y1, y2))  return true;
    }
    //nothing was hit, so return false
    return false;
  }

  //causes the monster to sink or rise
  void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }

  //hardcoding all of the timing info for the boss in millis
  void setTiming() {
    raiseTimes=new int[3];
    lowerTimes=new int[3];

    raiseTimes[0]=19120;
    lowerTimes[0]=25500;
    raiseTimes[1]=69500;
    lowerTimes[1]=81500;
    raiseTimes[2]=113900;
    lowerTimes[2]=150000;

    //using array list for now, because it makes it a bit easier to add
    ArrayList<Integer> whirlTimesList=new ArrayList<Integer>();
    ArrayList<Integer> whirlStrengthList=new ArrayList<Integer>();

    //FIRST APPEARANCE

    //set up drum hits
    int drumStartTime=69900;
    int drumEndTime=80000;
    int drumCycleTime=1565;  //how long it takes to go through a fulls et of drum hits
    //timing within the cycle
    int cycleTiming[]=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=400;
    cycleTiming[2]=566;
    cycleTiming[3]=800;
    cycleTiming[4]=1220;
    //power for each hit
    int cyclePower[]=new int[5];
    cyclePower[0]=150;
    cyclePower[1]=20;
    cyclePower[2]=20;
    cyclePower[3]=80;
    cyclePower[4]=80;

    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<cycleTiming.length; c++) {
        whirlTimesList.add(i+cycleTiming[c]);
        whirlStrengthList.add(cyclePower[c]);
      }
    }
    //one hit at the end of the segment
    whirlTimesList.add(81000);
    whirlStrengthList.add(200);


    //SECOND APPEARANCE
    drumStartTime=114080;  //115650;
    drumEndTime=137300;  //when the heavier drums kick in
    drumCycleTime=1575;  //how long it takes to go through a fulls et of drum hits

    //timing within the cycle
    cycleTiming=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=100;
    cycleTiming[2]=460;
    cycleTiming[3]=860;
    cycleTiming[4]=1260;
    //power for each hit
    cyclePower=new int[5];
    cyclePower[0]=60;
    cyclePower[1]=60;
    cyclePower[2]=40;
    cyclePower[3]=40;
    cyclePower[4]=40;

    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<cycleTiming.length; c++) {
        //need to sip the "Resua get down you'll be blown to simtherens bit" from 2:04.94 to 2:06.28
        if (i+cycleTiming[c]<124940 || i+cycleTiming[c]> 126280){
          whirlTimesList.add(i+cycleTiming[c]);
          whirlStrengthList.add(cyclePower[c]);
        }
      }
    }
    
    //add the hits during the Rensua bit
    int renHits[]=new int[4];
    renHits[0]=125430;
    renHits[1]=126000;
    renHits[2]=126180;
    renHits[3]=126350;
    int renIndex=0;
    //need to find where the other drums ended
    while(whirlTimesList.get(renIndex).intValue()<renHits[0])
      renIndex++;
    //now add them to the list of times starting at that spot
    for (int i=0; i<4; i++){
      whirlTimesList.add(renIndex+i,renHits[i]);
      whirlStrengthList.add(renIndex+i,100);
    }
    
    //drums have a steady kick leading into the heavier section
    int steadyStart=137580;
    int steadyEnd  =139010;
    int steadyPeriod  =210; 
    for (int i=steadyStart; i<steadyEnd; i+=steadyPeriod) {
      whirlTimesList.add(i);
      whirlStrengthList.add(70);
    }
    
    //heavier section
    drumStartTime=139400;  
    drumEndTime=149500;  //when the heavier drums kick in
    drumCycleTime=1579;  //how long it takes to go through a fulls et of drum hits

    //timing within the cycle MAYBE WE ONLY NEED 3, NOT ALL 5
    cycleTiming=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=372;
    cycleTiming[2]=690;
    cycleTiming[3]=998;
    cycleTiming[4]=1684;
    //power for each hit
    cyclePower=new int[5];
    cyclePower[0]=40;
    cyclePower[1]=40;
    cyclePower[2]=40;
    cyclePower[3]=40;
    cyclePower[4]=40;
    
    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<3; c++) {
        whirlTimesList.add(i+cycleTiming[c]);
        whirlStrengthList.add(cyclePower[c]);
      }
    }
    
    

    //convert the arraylists to good old arrays after
    whirlTimes=new int[whirlTimesList.size()];
    whirlStrength=new int[whirlStrengthList.size()];
    for (int i=0; i<whirlTimesList.size(); i++) {
      whirlTimes[i]= whirlTimesList.get(i).intValue();
      whirlStrength[i]= whirlStrengthList.get(i).intValue();
    }
  }

  //advances all timers to this point in the song
  void skipTo(int time) {
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;

    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;

    for (int i=0; i<whirlTimes.length;i++)
      if (time > whirlTimes[i])  nextWhirlTime=i;
  }
}

