class Ship {
  float x, y;  //lcoation

  //movment
  float moveSpeed=5;  //speed (no surprise there)
  float padding=50;    //how close to the edges to let the ship get

  //bobbing
  float bobAngle=0;
  float bobSpeed=0.005;
  float bobMax=0.1;  //angle in radians that the boat can bob

  //capping the fire rate
  int shotTime=20;  //how many frames between shots
  int shotTimer=0;  //the timer

  //getting hit
  float hitX=0;    //offset from regular x
  float hitY=30;   //offset from regular y
  float hitW=80;
  float hitH=30;
  int hitTime=30;  //how many frames to show red upon being hit
  int hitTimer=0;

  //graphics
  PImage pic; //pic
  
  //is the ship dead?
  boolean isDead;
  int sinkingMask[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored
  float sinkingY;      //this should start off being the height of the ship
  float sinkingSpeed=4;

  Ship() {
    x=100;
    y=height/2; 

    isDead=false;

    //load in the pic
    pic=loadImage("data/pics/ship.png");
    sinkingMask=new int[pic.width*pic.height];
    sinkingY=pic.height;
    
    //set which pixels can be transparent
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      color col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }
    
  }

  void update() {
    shotTimer++;  //increase the shot timer
    hitTimer--;  //reduce the hit timer

    //keep the ship in bounds
    if (x<padding)        x=padding;
    if (x>width-padding)  x=width-padding;
    if (y<padding)        y=padding;
    if (y>height-padding) y=height-padding;

    //have the boat bob a bit
    bobAngle+=bobSpeed;
    if (bobAngle>bobMax || bobAngle<-bobMax)
      bobSpeed*=-1;
      
    //if it is dead, sink the ship using the mask
    if (isDead){
      sinkingY-=sinkingSpeed;
      for (int c=0; c<pic.width; c++){
        for (int r=0; r<pic.height; r++){
          int pos=r*pic.width+c;
          
          //assume we want to draw this pixel
          sinkingMask[pos]=255;
          
          //check if this is one of the values that must be transparent
          if (maskTrans[pos])  sinkingMask[pos]=0;
          
          //if this is below the sinking level, make it transparent
          if (r>sinkingY)  sinkingMask[pos]=0;
        }
      }
    }
  }

  void draw() {
    imageMode(CENTER); 

    //rotate a bit for bob

    pushMatrix();
    translate(x, y);
    //if the player is dying, move the ship down
    if (isDead) translate(0,pic.height-sinkingY);
    
    //mask some of the image if it is sinking
    if (isDead)  pic.mask(sinkingMask);

    //translate(0,pic.width/2);
    rotate(bobAngle);
    //translate(0,-pic.width/2);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    image(pic, 0, 0);
    popMatrix();

    //draw hitRange
//    rectMode(CENTER);
//    noFill();
//    stroke(0, 255, 0);
//    rect(x+hitX, y+hitY, hitW, hitH);
  }

  //moves the ship
  void moveShip(int horzDir, int vertDir) {
    if (!isDead){
      x+=moveSpeed*horzDir;
      y+=moveSpeed*vertDir;
    }
  }

  //handles shooting a cannonball
  //actual bullet spawn is handled in main, this handles other effects
  void shoot() {
    shotTimer=0;
  }

  //called when the ship gets hurt
  void hit() {
    if (!isDead){
      hitTimer=hitTime;
      if (playerHealth<=0)
        isDead=true;
    }
      
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }
}

