import ddf.minim.*;
import ddf.minim.signals.*;
import ddf.minim.analysis.*;
import ddf.minim.effects.*;

//player
Ship ship;

//starting the gane
PImage startPic;
int startScreenTimer=1100;

//ending the game
boolean gameOver;
float endGameTimer=0;
float endGameSpeed=0.5;
PImage losePic;
PImage winPic;

//bosses
BossShooter bossShooter;
BossWhirlPool bossWhirlPool;
BossCharger bossCharger;
Finisher finisher;

//stats
float playerHealthStart=26;
float playerHealth=playerHealthStart;
float bossHealthStart=150;
float bossHealth=bossHealthStart;

//background ocean
Backdrop backdrop;

//controls. Keeps track of which keys are down
boolean[] keysDown=new boolean[127];

//bullets
ArrayList bullets;
boolean lastBulletInFront;  //was the last bullet in front of the ship? Used to alternate shots
ArrayList bossBullets;

//drawing the health
PImage playerHealthFrame;
PImage bossHealthFrame;
PImage healthBarPic;
//masking arrays to hide the health bar as health goes down
int playerHealthMask[];
int bossHealthMask[];

//audio
Minim minim;
AudioPlayer song;
AudioPlayer deathSong;
AudioPlayer winSong;

int spaceTimer=0;

void setup() {
  smooth();
  size(1200, 800);
  frameRate(60);
  
  gameOver=false;
  
  //get the bullets alternating
  lastBulletInFront=false;
  
  //load images for the health bar
  playerHealthFrame=loadImage("data/pics/health/playerHealthFrame.png");
  bossHealthFrame=loadImage("data/pics/health/monsterHealthFrame.png");
  healthBarPic=loadImage("data/pics/health/healthBar.png");
  playerHealthMask=new int[healthBarPic.width*healthBarPic.height];
  bossHealthMask=new int[healthBarPic.width*healthBarPic.height];

  //set up game elements
  ship=new Ship();
  bossShooter=new BossShooter();
  bossWhirlPool=new BossWhirlPool();
  bossCharger=new BossCharger();
  finisher=new Finisher();
  backdrop=new Backdrop();
  bullets=new ArrayList();
  bossBullets=new ArrayList();
  
  //screens
  startPic=loadImage("screens/intro.png");
  losePic=loadImage("screens/loseScreen.jpg");
  winPic=loadImage("screens/winScreen.png");
  
  //set up the song
  minim = new Minim(this);
  song = minim.loadFile("audio/sonsofthewaves.mp3");
  deathSong=minim.loadFile("audio/deathSong.mp3");
  winSong=minim.loadFile("audio/winSong.mp3");
  song.play();
  
  int skipTime=00;
  song.skip(skipTime);
  bossShooter.skipTo(skipTime);
  bossWhirlPool.skipTo(skipTime);
  bossCharger.skipTo(skipTime);
  finisher.skipTo(skipTime);
  

}

void update() {
  //if the intro screen is still up, more that along
  startScreenTimer-=2;
  
//check the controls
  //left
  if (keysDown[37] || keysDown[65])  ship.moveShip(-1, 0);
  //right
  if (keysDown[39] || keysDown[68])  ship.moveShip(1, 0);
  //up
  if (keysDown[38] || keysDown[87])  ship.moveShip(0, -1);
  //down
  if (keysDown[40] || keysDown[83])  ship.moveShip(0, 1);
  //space to shoot
  if (keysDown[32])  shootBullet();


  backdrop.update();
  ship.update();
  bossShooter.update(song.position());
  bossCharger.update(song.position());
  bossWhirlPool.update(song.position());
  if (!gameOver)  //dont' let the finisher happen if the player just won
    finisher.update(song.position());
  
  //check if shooting boss fired
  if (bossShooter.shooting){
    shootBossBullet();
  }

  //update bullets
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    b.update();

    //see if it went offscreen
    if (b.x>width+100) {
      bullets.remove(i);
    }
    //see if it hit a boss
    //shooter boss
    else if (bossShooter.active && bossShooter.hitCheck(b.x, b.y, b.hitSize)) {
      bossShooter.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
    //whirlpool boss
    else if (bossWhirlPool.active && bossWhirlPool.hitCheck(b.x, b.y, b.hitSize)) {
      bossWhirlPool.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
    //charging boss
    else if (bossCharger.active && bossCharger.hitCheck(b.x, b.y, b.hitSize)) {
      bossCharger.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
  }

  //update bossBullets
  for (int i=0; i<bossBullets.size(); i++) {
    BossBullet b=(BossBullet)bossBullets.get(i);
    b.update();
    //see if they went off screen
    if (b.x<-100) {
      bossBullets.remove(i);
    }
    //see if they hit the player
    if (ship.hitCheck(b.x, b.y, b.hitSize) && !ship.isDead && ship.hitTimer<0) {
      ship.hit();
      playerHealth-=b.damage;
      //kill the bullet
      bossBullets.remove(i);
    }
  }

  //if the whirlpool momster is active, check if any whirlpools hit the player
  if (bossWhirlPool.active) {
    if (bossWhirlPool.whirlPoolHitCheck(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) && ship.hitTimer<0) {
      ship.hit();
      playerHealth-=bossWhirlPool.whirlpoolDamage;
    }
  }

  //check if the player touched a boss
  if (ship.hitTimer<0 && (
  bossCharger.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) ||
    bossShooter.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) ||
    bossWhirlPool.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2))) {
    ship.hit(); 
    playerHealth-=bossCharger.damage;
  }
  
  //check if the finisher just did its j=job
  if (finisher.killShip && !ship.isDead){
    ship.hit();
    playerHealth-=9999; 
  }
  
  //check if the player died
  if (ship.isDead && !gameOver)  playerLose();
  //check if the boss has dies
  if (bossHealth<=0 && !gameOver)   playerWin();
  
  if (gameOver)  endGameTimer+=endGameSpeed;
}

void draw() {
  background(0);  //clear the screen
  update();  //do everything in update first

  //draw the ocean
  backdrop.draw();

  //draw bullets in behind the ship
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    if (!b.inFront)
      b.draw();
  }

  bossWhirlPool.draw();
  bossShooter.draw();
  bossCharger.draw();
  ship.draw(); 
  finisher.draw();
  
  //draw bossBullets
  for (int i=0; i<bossBullets.size(); i++) {
    BossBullet b=(BossBullet)bossBullets.get(i);
    b.draw();
  }

  //draw bullets in front of the ship
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    if (b.inFront)
      b.draw();
  }
  
  drawHealth();
  
  //flash you win of the player won
  if (gameOver && !ship.isDead){
    //eventualy fade in a black box behind
   float boxFadeIn=180;
   if(endGameTimer>boxFadeIn){
     rectMode(CENTER);
     fill(0,0,0, (endGameTimer-boxFadeIn)*2);
     rect(width/2,height/2,width+3,height+3);  //keep the border of the box just off screen
   }
   
   //show the text
   tint(random(255),random(255),random(255),255);
   image(winPic,width/2,height/2,width,height); 
   
   
  }
  
  //fade to black if the player died
  if (ship.isDead){
   tint(255,255,255,endGameTimer-10);
   image(losePic,width/2,height/2,width,height); 
  }
  
  //show the intro screen if it is still up
  if (startScreenTimer>0){
    tint(255,255,255,startScreenTimer);  //always keep some transparency
    image(startPic,width/2,height/2);
  }

}


//draw the health info
void drawHealth(){
  //placement info
  float barX=310;
  float frameY=50;
  float barY=frameY+15;  //center of bar needs to be shown a bit lower than the frame
  imageMode(CENTER); 
  
  //set the mask for the player health bar
  int playerMaskCutOff=(int)map(playerHealth,0,playerHealthStart,0,healthBarPic.width);
  //go through and set the pixels n the array
  for (int x=0; x<healthBarPic.width; x++){
    //check if this column should be drawn
    int alphaVal=0;
    if (x<playerMaskCutOff)  alphaVal=255;
    
    for (int y=0; y<healthBarPic.height; y++){
      int pos=y*healthBarPic.width+x;
      playerHealthMask[pos]=alphaVal;
    }
  }
  
  //set the mask for the boss health bar
  int bossMaskCutOff=(int)map(bossHealth,0,bossHealthStart,healthBarPic.width,0);
  //go through and set the pixels n the array
  for (int x=0; x<healthBarPic.width; x++){
    //check if this column should be drawn
    int alphaVal=0;
    if (x>bossMaskCutOff)  alphaVal=255;  //cut off in the opposite direciton because this health bar moves the opposite way
    
    for (int y=0; y<healthBarPic.height; y++){
      int pos=y*healthBarPic.width+x;
      bossHealthMask[pos]=alphaVal;
    }
  }
  
  //draw the health bars
  //player bar
  tint(214,190,124);
  healthBarPic.mask(playerHealthMask);
  image(healthBarPic, barX, barY);  
  //boss bar  
  tint(217,124,119);
  healthBarPic.mask(bossHealthMask);
  image(healthBarPic, width-barX, barY);  
  
  //make the frames be red if the player or monster just got hit
  //monster frame
  if ( (bossShooter.hitTimer>0 || bossCharger.hitTimer>0 || bossWhirlPool.hitTimer>0) && frameCount%10>5) tint(255,0,0);
  else             tint(255,255,255);
  image(bossHealthFrame, width-barX, frameY);
  //player frame
  if (ship.hitTimer>0 && frameCount%10>5)  tint(255,0,0);
  else             tint(255,255,255);
  image(playerHealthFrame, barX, frameY);
  
}

//if the timer is ready, shoots a cannonball from the ship
void shootBullet() {
  if (ship.shotTimer>=ship.shotTime && !ship.isDead) {
    Bullet newBullet=new Bullet(ship.x, ship.y, lastBulletInFront);
    bullets.add(newBullet);
    ship.shoot();
    lastBulletInFront=!lastBulletInFront;
  }
}

void shootBossBullet() {
  bossShooter.shooting=false;
  //get the angle the shot should be at
  float shotAngle;
  if (bossShooter.shootHigh)
    shotAngle=random((3*PI)/4, PI);
  else
    shotAngle=random(PI, (5*PI)/4);

  //spawn a bullet at that bosse's location
  BossBullet newBullet=new BossBullet(bossShooter.x+bossShooter.xShotOffset, bossShooter.y+bossShooter.yShotOffset, shotAngle);
  bossBullets.add(newBullet);
}


void playerLose(){
 gameOver=true;
 song.mute();
 deathSong.play(); 
}

void playerWin(){
 gameOver=true;
 song.mute();
 winSong.play(); 
 //have the bosses retreat
 bossShooter.raise(-1);
 bossWhirlPool.raise(-1);
 bossCharger.raise(-1);
}

//listen for keyboard input
void keyPressed() {
  keysDown[keyCode]=true;  //mark that they key was pressed down

  //testing

//  if (key=='1')  bossShooter.raise(1);
//  if (key=='2')  bossShooter.raise(-1);
//  if (key=='q')  bossWhirlPool.raise(1);
//  if (key=='w')  bossWhirlPool.raise(-1);
//  if (key=='a')  bossCharger.raise(1);
//  if (key=='s')  bossCharger.raise(-1);
//
//  if (key=='3')  shootBossBullet();
//  if (key=='e')  bossWhirlPool.setPools(200);
//  if (key=='r')  bossWhirlPool.setPools(250);
//  if (key=='t')  bossWhirlPool.setPools(300);
  
}
void keyReleased() {
  keysDown[keyCode]=false;  //mark that they key was released
}


//kill the sound on close
void stop() {
  song.close();
  minim.stop();
  super.stop();
}


