import processing.core.*; 
import processing.xml.*; 

import ddf.minim.*; 
import ddf.minim.signals.*; 
import ddf.minim.analysis.*; 
import ddf.minim.effects.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class SonsOfTheWaves extends PApplet {






//player
Ship ship;

//starting the gane
PImage startPic;
int startScreenTimer=1100;

//ending the game
boolean gameOver;
float endGameTimer=0;
float endGameSpeed=0.5f;
PImage losePic;
PImage winPic;

//bosses
BossShooter bossShooter;
BossWhirlPool bossWhirlPool;
BossCharger bossCharger;
Finisher finisher;

//stats
float playerHealthStart=26;
float playerHealth=playerHealthStart;
float bossHealthStart=150;
float bossHealth=bossHealthStart;

//background ocean
Backdrop backdrop;

//controls. Keeps track of which keys are down
boolean[] keysDown=new boolean[127];

//bullets
ArrayList bullets;
boolean lastBulletInFront;  //was the last bullet in front of the ship? Used to alternate shots
ArrayList bossBullets;

//drawing the health
PImage playerHealthFrame;
PImage bossHealthFrame;
PImage healthBarPic;
//masking arrays to hide the health bar as health goes down
int playerHealthMask[];
int bossHealthMask[];

//audio
Minim minim;
AudioPlayer song;
AudioPlayer deathSong;
AudioPlayer winSong;

int spaceTimer=0;

public void setup() {
  smooth();
  size(1200, 800);
  frameRate(60);
  
  gameOver=false;
  
  //get the bullets alternating
  lastBulletInFront=false;
  
  //load images for the health bar
  playerHealthFrame=loadImage("data/pics/health/playerHealthFrame.png");
  bossHealthFrame=loadImage("data/pics/health/monsterHealthFrame.png");
  healthBarPic=loadImage("data/pics/health/healthBar.png");
  playerHealthMask=new int[healthBarPic.width*healthBarPic.height];
  bossHealthMask=new int[healthBarPic.width*healthBarPic.height];

  //set up game elements
  ship=new Ship();
  bossShooter=new BossShooter();
  bossWhirlPool=new BossWhirlPool();
  bossCharger=new BossCharger();
  finisher=new Finisher();
  backdrop=new Backdrop();
  bullets=new ArrayList();
  bossBullets=new ArrayList();
  
  //screens
  startPic=loadImage("screens/intro.png");
  losePic=loadImage("screens/loseScreen.jpg");
  winPic=loadImage("screens/winScreen.png");
  
  //set up the song
  minim = new Minim(this);
  song = minim.loadFile("audio/sonsofthewaves.mp3");
  deathSong=minim.loadFile("audio/deathSong.mp3");
  winSong=minim.loadFile("audio/winSong.mp3");
  song.play();
  
  int skipTime=00;
  song.skip(skipTime);
  bossShooter.skipTo(skipTime);
  bossWhirlPool.skipTo(skipTime);
  bossCharger.skipTo(skipTime);
  finisher.skipTo(skipTime);
  

}

public void update() {
  //if the intro screen is still up, more that along
  startScreenTimer-=2;
  
//check the controls
  //left
  if (keysDown[37] || keysDown[65])  ship.moveShip(-1, 0);
  //right
  if (keysDown[39] || keysDown[68])  ship.moveShip(1, 0);
  //up
  if (keysDown[38] || keysDown[87])  ship.moveShip(0, -1);
  //down
  if (keysDown[40] || keysDown[83])  ship.moveShip(0, 1);
  //space to shoot
  if (keysDown[32])  shootBullet();


  backdrop.update();
  ship.update();
  bossShooter.update(song.position());
  bossCharger.update(song.position());
  bossWhirlPool.update(song.position());
  if (!gameOver)  //dont' let the finisher happen if the player just won
    finisher.update(song.position());
  
  //check if shooting boss fired
  if (bossShooter.shooting){
    shootBossBullet();
  }

  //update bullets
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    b.update();

    //see if it went offscreen
    if (b.x>width+100) {
      bullets.remove(i);
    }
    //see if it hit a boss
    //shooter boss
    else if (bossShooter.active && bossShooter.hitCheck(b.x, b.y, b.hitSize)) {
      bossShooter.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
    //whirlpool boss
    else if (bossWhirlPool.active && bossWhirlPool.hitCheck(b.x, b.y, b.hitSize)) {
      bossWhirlPool.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
    //charging boss
    else if (bossCharger.active && bossCharger.hitCheck(b.x, b.y, b.hitSize)) {
      bossCharger.hit();
      bossHealth-=b.damage;
      //kill the bullet
      bullets.remove(i);
    }
  }

  //update bossBullets
  for (int i=0; i<bossBullets.size(); i++) {
    BossBullet b=(BossBullet)bossBullets.get(i);
    b.update();
    //see if they went off screen
    if (b.x<-100) {
      bossBullets.remove(i);
    }
    //see if they hit the player
    if (ship.hitCheck(b.x, b.y, b.hitSize) && !ship.isDead && ship.hitTimer<0) {
      ship.hit();
      playerHealth-=b.damage;
      //kill the bullet
      bossBullets.remove(i);
    }
  }

  //if the whirlpool momster is active, check if any whirlpools hit the player
  if (bossWhirlPool.active) {
    if (bossWhirlPool.whirlPoolHitCheck(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) && ship.hitTimer<0) {
      ship.hit();
      playerHealth-=bossWhirlPool.whirlpoolDamage;
    }
  }

  //check if the player touched a boss
  if (ship.hitTimer<0 && (
  bossCharger.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) ||
    bossShooter.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2) ||
    bossWhirlPool.isBoxTouching(ship.x+ship.hitX-ship.hitW/2, ship.x+ship.hitX+ship.hitW/2, ship.y+ship.hitY-ship.hitH/2, ship.y+ship.hitY+ship.hitH/2))) {
    ship.hit(); 
    playerHealth-=bossCharger.damage;
  }
  
  //check if the finisher just did its j=job
  if (finisher.killShip && !ship.isDead){
    ship.hit();
    playerHealth-=9999; 
  }
  
  //check if the player died
  if (ship.isDead && !gameOver)  playerLose();
  //check if the boss has dies
  if (bossHealth<=0 && !gameOver)   playerWin();
  
  if (gameOver)  endGameTimer+=endGameSpeed;
}

public void draw() {
  background(0);  //clear the screen
  update();  //do everything in update first

  //draw the ocean
  backdrop.draw();

  //draw bullets in behind the ship
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    if (!b.inFront)
      b.draw();
  }

  bossWhirlPool.draw();
  bossShooter.draw();
  bossCharger.draw();
  ship.draw(); 
  finisher.draw();
  
  //draw bossBullets
  for (int i=0; i<bossBullets.size(); i++) {
    BossBullet b=(BossBullet)bossBullets.get(i);
    b.draw();
  }

  //draw bullets in front of the ship
  for (int i=0; i<bullets.size(); i++) {
    Bullet b=(Bullet)bullets.get(i);
    if (b.inFront)
      b.draw();
  }
  
  drawHealth();
  
  //flash you win of the player won
  if (gameOver && !ship.isDead){
    //eventualy fade in a black box behind
   float boxFadeIn=180;
   if(endGameTimer>boxFadeIn){
     rectMode(CENTER);
     fill(0,0,0, (endGameTimer-boxFadeIn)*2);
     rect(width/2,height/2,width+3,height+3);  //keep the border of the box just off screen
   }
   
   //show the text
   tint(random(255),random(255),random(255),255);
   image(winPic,width/2,height/2,width,height); 
   
   
  }
  
  //fade to black if the player died
  if (ship.isDead){
   tint(255,255,255,endGameTimer-10);
   image(losePic,width/2,height/2,width,height); 
  }
  
  //show the intro screen if it is still up
  if (startScreenTimer>0){
    tint(255,255,255,startScreenTimer);  //always keep some transparency
    image(startPic,width/2,height/2);
  }

}


//draw the health info
public void drawHealth(){
  //placement info
  float barX=310;
  float frameY=50;
  float barY=frameY+15;  //center of bar needs to be shown a bit lower than the frame
  imageMode(CENTER); 
  
  //set the mask for the player health bar
  int playerMaskCutOff=(int)map(playerHealth,0,playerHealthStart,0,healthBarPic.width);
  //go through and set the pixels n the array
  for (int x=0; x<healthBarPic.width; x++){
    //check if this column should be drawn
    int alphaVal=0;
    if (x<playerMaskCutOff)  alphaVal=255;
    
    for (int y=0; y<healthBarPic.height; y++){
      int pos=y*healthBarPic.width+x;
      playerHealthMask[pos]=alphaVal;
    }
  }
  
  //set the mask for the boss health bar
  int bossMaskCutOff=(int)map(bossHealth,0,bossHealthStart,healthBarPic.width,0);
  //go through and set the pixels n the array
  for (int x=0; x<healthBarPic.width; x++){
    //check if this column should be drawn
    int alphaVal=0;
    if (x>bossMaskCutOff)  alphaVal=255;  //cut off in the opposite direciton because this health bar moves the opposite way
    
    for (int y=0; y<healthBarPic.height; y++){
      int pos=y*healthBarPic.width+x;
      bossHealthMask[pos]=alphaVal;
    }
  }
  
  //draw the health bars
  //player bar
  tint(214,190,124);
  healthBarPic.mask(playerHealthMask);
  image(healthBarPic, barX, barY);  
  //boss bar  
  tint(217,124,119);
  healthBarPic.mask(bossHealthMask);
  image(healthBarPic, width-barX, barY);  
  
  //make the frames be red if the player or monster just got hit
  //monster frame
  if ( (bossShooter.hitTimer>0 || bossCharger.hitTimer>0 || bossWhirlPool.hitTimer>0) && frameCount%10>5) tint(255,0,0);
  else             tint(255,255,255);
  image(bossHealthFrame, width-barX, frameY);
  //player frame
  if (ship.hitTimer>0 && frameCount%10>5)  tint(255,0,0);
  else             tint(255,255,255);
  image(playerHealthFrame, barX, frameY);
  
}

//if the timer is ready, shoots a cannonball from the ship
public void shootBullet() {
  if (ship.shotTimer>=ship.shotTime && !ship.isDead) {
    Bullet newBullet=new Bullet(ship.x, ship.y, lastBulletInFront);
    bullets.add(newBullet);
    ship.shoot();
    lastBulletInFront=!lastBulletInFront;
  }
}

public void shootBossBullet() {
  bossShooter.shooting=false;
  //get the angle the shot should be at
  float shotAngle;
  if (bossShooter.shootHigh)
    shotAngle=random((3*PI)/4, PI);
  else
    shotAngle=random(PI, (5*PI)/4);

  //spawn a bullet at that bosse's location
  BossBullet newBullet=new BossBullet(bossShooter.x+bossShooter.xShotOffset, bossShooter.y+bossShooter.yShotOffset, shotAngle);
  bossBullets.add(newBullet);
}


public void playerLose(){
 gameOver=true;
 song.mute();
 deathSong.play(); 
}

public void playerWin(){
 gameOver=true;
 song.mute();
 winSong.play(); 
 //have the bosses retreat
 bossShooter.raise(-1);
 bossWhirlPool.raise(-1);
 bossCharger.raise(-1);
}

//listen for keyboard input
public void keyPressed() {
  keysDown[keyCode]=true;  //mark that they key was pressed down

  //testing

//  if (key=='1')  bossShooter.raise(1);
//  if (key=='2')  bossShooter.raise(-1);
//  if (key=='q')  bossWhirlPool.raise(1);
//  if (key=='w')  bossWhirlPool.raise(-1);
//  if (key=='a')  bossCharger.raise(1);
//  if (key=='s')  bossCharger.raise(-1);
//
//  if (key=='3')  shootBossBullet();
//  if (key=='e')  bossWhirlPool.setPools(200);
//  if (key=='r')  bossWhirlPool.setPools(250);
//  if (key=='t')  bossWhirlPool.setPools(300);
  
}
public void keyReleased() {
  keysDown[keyCode]=false;  //mark that they key was released
}


//kill the sound on close
public void stop() {
  song.close();
  minim.stop();
  super.stop();
}


class Backdrop {
  float x;
  float speed;
  PImage pic;
  
  int numTiles;  //how many times the image must be drawn to stretch across the screen
  
  Backdrop(){
    x=0;  //x is used to cycle the image over and over
    speed=2;  //how fast the sea moves 
    
    numTiles=4;
    
    //load the image
    pic=loadImage("data/pics/waterTile.jpg");
  }
  
  public void update(){
    x-=speed;
    if (x<-width)  x+=pic.width;
  }
  
  public void draw(){
    tint(255,255,255);
    imageMode(CORNER);
    for (int i=0; i<numTiles; i++)
      image(pic,x+i*pic.width,0);
  }
}

class BossBullet{
  float x,y;  //location
  float angle;
  
  float hitSize=30;
  float damage=2;
  
  //how far from the center of the ship to start it
  float inFrontOffset=35;  
  float inBackOffset=25;
  
  float speed=8;  //how fast it moves
  float xVel;
  float yVel;
  
  PImage pic;   //image of the cannon ball
  float picOffsetX=35;  //how much to move the image over to keep the front in the center
  
  boolean killMe;  //flag becomes true if the cannon ball is off screen
  
 
  BossBullet(float _x, float _y, float _angle){
    x=_x;
    y=_y;
    angle=_angle;
    
    xVel=cos(angle)*speed;
    yVel=sin(angle)*speed;
    
    killMe=false;
    
    //load the image
    pic=loadImage("data/pics/bossShot.png");
   
  } 
  
  public void update(){
    //move the ball
    x+=xVel;
    y+=yVel;
    //check if it is offscreen
    if (x<0)  killMe=true;
    
    
  }
  
  public void draw(){
    imageMode(CENTER); 
    pushMatrix();
    translate(x,y);
    rotate(angle+PI);
    translate(picOffsetX,0);  //move the image over, so the hit box is on the front
    tint(255,255,255);
    image(pic,0,0);
    popMatrix();
   
    //draw the hit box 
//    noFill();
//    stroke(255, 0, 0);
//    ellipse(x,y,hitSize,hitSize);
  }
  
  
}
class BossCharger {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //moving
  //collection of when the monster should charge toward the player
  int chargeTime=30;  //how long it should take to reach the player in frames
  int retreatTime=30;
  float chargeX=180;  //once the boss hits this x value, the charge is done

  //returning from a charge
  float waitYPadding=100;  //how far from the top and bottom it can be
  float waitX=width-180;  //goal x Point when returning from a charge
  int waitTime=50;        //NOT USED aproximate number of frames it should be waiting there 

  //these are set when a charge starts or ends
  float xVel=0;
  float yVel=0;  

  //getting hit
  float hitX=-70;    //offset from regular x
  float hitY=30;   //offset from regular y
  float hitW=140;
  float hitH=50;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showing the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=2;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=50;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  float damage=3;  //damage form touching the boss

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int moveTimes[];
  int nextMoveTime=0;

  BossCharger() {
    x=waitX;
    y=height/2; 

    pic=loadImage("data/pics/chargeMonster.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      int col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    setTimes();
  }

  public void update(int time) {

    if (!gameOver)
      checkTiming(time);

    //if the beats is over the line, start retreating
    if (x<chargeX) {
      //claculate a random y point to try and return to
      float targetY=random(waitYPadding, height-waitYPadding);

      //set the velocities to try to hit that point
      xVel=(waitX-x)/retreatTime;
      yVel=(targetY-y)/retreatTime;
    }

    //if it has crossed the wait line, just wait
    if (x>waitX) {
      x=waitX;
      xVel=0;
      yVel=0;
    }

    x+=xVel;
    y+=yVel;


    //if this is at all visible, it is active
    active=raisingY>0;

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;

      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //adjust the mask for the image
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1f, millis()*0.001f)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //lower the hitTimer
    hitTimer--;
  }

  public void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //see if it is time to charge
    if (nextMoveTime<moveTimes.length) {
      if (time>moveTimes[nextMoveTime]) {
        //charge at player 
        //calculate the speed to move to reahc player pos in the given amount of time
        xVel=(ship.x-x)/chargeTime;
        yVel=(ship.y-y)/chargeTime;

        nextMoveTime++;  //increase to the next time in the arra
      }
    }
  }

  public void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

  }

  //called when the boss gets hurt
  public void hit() {
    hitTimer=hitTime;
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  public boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }


  //causes the monster to sink or rise
  public void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  public boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }
  
   //hard code the times that this thing needs to charge
  public void setTimes() {
    raiseTimes=new int[2];
    lowerTimes=new int[2];

    raiseTimes[0]=16050;
    lowerTimes[0]=25500;
    raiseTimes[1]=81500;
    lowerTimes[1]=107600;
    
    //use an arryalist for now because it is easier to ad to
    ArrayList<Integer> moveTimesList=new ArrayList<Integer>();
    
    moveTimesList.add(82570);
    moveTimesList.add(85720);
    moveTimesList.add(88910);
    moveTimesList.add(92020);
    moveTimesList.add(95200);
    moveTimesList.add(98160);
    
    moveTimesList.add(101330);
    moveTimesList.add(102100);
    moveTimesList.add(102870);
    moveTimesList.add(103680);
    
    moveTimesList.add(104450);
    moveTimesList.add(105260);
    
    moveTimesList.add(99999999);  //put the last charge at a time that can never happen
    //convert the arraylist to good old array
    moveTimes=new int[moveTimesList.size()];
    for (int i=0; i<moveTimesList.size(); i++){
      moveTimes[i]= moveTimesList.get(i).intValue();
    }
    
  }
  
  //advances all timers to this point in the song
  public void skipTo(int time) {
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;

    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;

    for (int i=0; i<moveTimes.length;i++)
      if (time > moveTimes[i])  nextMoveTime=i;
  }
  
}

class BossShooter {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //how much to slow down the values coming into the noise function
  float noiseSpeedX=0.00018f; 
  float noiseSpeedY=0.0002f;  
  //where in the board it will be
  float xRange=200;
  float yRange=300;
  float xCenter=width-400;
  float yCenter=height/2;
  //keep it moving up and down
  //y locaiton from noise will be centerred around this
  float ySpeed=0.0002f;
  float yPadding=100;  //keep the monster on screen

  //shooting
  float yShotOffset=-30;
  float xShotOffset=-30;
  boolean shootHigh=true;  //boss should alternate shooting high and low
  boolean shooting=false;  //flag for main to shoot a bullet

  //getting hit
  float hitX=25;    //offset from regular x
  float hitY=10;   //offset from regular y
  float hitW=40;
  float hitH=100;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showin the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=4;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=10;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  float damage=3;  //damage from touching the beast

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int shotTimes[];
  int nextShotTime=0;

  BossShooter() {
    x=xCenter;
    y=yCenter; 

    pic=loadImage("data/pics/shooterBoss.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      int col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    setTiming();
  }

  public void update(int time) {
    //keep it moving up and down
    yCenter=height/2+sin(millis()*ySpeed)*yPadding;

    x=xCenter+noise((float)millis()*noiseSpeedX, 1)*xRange*2-xRange;
    y=yCenter+noise((float)millis()*noiseSpeedY, 1)*yRange*2-yRange;
    //the 1 just serves to seperate if from the noise being generated for the other monsters

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;
      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //if this is at all visible, it is active
    active=raisingY>0;

    //adjust the mask for the beast
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1f, millis()*0.001f)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //lower the hitTimer
    hitTimer--;

    if (!gameOver)
      checkTiming(time);  //see if anything needs to happen based on timing
  }

  public void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //time to shoot?
    if (nextShotTime<shotTimes.length) {
      if (time>shotTimes[nextShotTime]) {
        shoot();
        nextShotTime++;
      }
    }
  }

  public void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

  }

  //does the animaiton for shooting
  //firing the bullet is handled in main
  public void shoot() { 
    shootHigh=!shootHigh;
    shooting=true;
  }

  //called when the boss gets hurt
  public void hit() {
    hitTimer=hitTime;
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  public boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }

  //causes the monster to sink or rise
  public void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  public boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }

  //hardcoding all of the timing info for the boss in millis
  public void setTiming() {
    raiseTimes=new int[3];
    lowerTimes=new int[3];

    raiseTimes[0] =12900;
    lowerTimes[0] =25500;
    raiseTimes[1] =31400;
    lowerTimes[1] =69500;
    raiseTimes[2]=138800;
    lowerTimes[2]=162800;

    //fill up the shooting array list then convert to normal array
    ArrayList<Integer> shotTimesList=new ArrayList<Integer>();
    
    //first segment starts at 31.84 and occurs every 0.4 seconds
    //THIS PART WAS CODED DIFFERENTLY THAN THE OTHERS BECAUSE I DID IT FIRST
    int time=69500-31840;
    int shotSpacing=400;
    int numShots=time/shotSpacing;
    for (int i=0; i<numShots; i++)
      shotTimesList.add(31840+i*shotSpacing);
      
      
    //second segment
    int startTime=139200;
    int endTime=  150250;
    shotSpacing=400;
    
    for (int i=startTime; i<endTime; i+=shotSpacing)
      shotTimesList.add(i);
      
    //quick series of guitar hits then a  pause
    startTime=endTime;
    endTime=151130;
    shotSpacing=100;
    for (int i=startTime; i<endTime; i+=shotSpacing)
      shotTimesList.add(i);
      
      
    //a few series of 5 notes
    //timing within the cycle
    startTime=151900;  //115650;
    endTime=159000;  //when the guitars end
    int cycleTime=3181;  //how long it takes to go through a full cycle
    
    int cycleTiming[]=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=1331;
    cycleTiming[2]=1493;
    cycleTiming[3]=2190;
    cycleTiming[4]=2337;
    
    for (int i=startTime; i<endTime; i+=cycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<5; c++) {
        shotTimesList.add(i+cycleTiming[c]);
      }
    }
    
    //2 more to finish it out
    shotTimesList.add(161360);
    shotTimesList.add(161460);
    
      
      
    //convert the arraylist to good old array
    shotTimes=new int[shotTimesList.size()];
    for (int i=0; i<shotTimesList.size(); i++){
      shotTimes[i]= shotTimesList.get(i).intValue();
    }
  }
  
  //advances all timers to this point in the song
  public void skipTo(int time){
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;
    
    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;
      
    for (int i=0; i<shotTimes.length; i++)
      if (time > shotTimes[i])  nextShotTime=i;
  }
  
}

class BossWhirlPool {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play

  //how much to slow down the values coming into the noise function
  float noiseSpeedX=0.0003f; 
  float noiseSpeedY=0.0002f;  
  //where in the board it will be
  float xRange=300;
  float yRange=300;
  float xCenter=width-400;
  float yCenter=height/2;
  //keep it moving up and down
  //y locaiton from noise will be centerred around this
  float ySpeed=0.0002f;
  float yPadding=100;  //keep the monster on screen
  //x location from noise will be centerred aorund this 
  float xSpeed=0.0007f;
  float xPadding=100;  //keep the monster on screen
  float xShift=-150;  //how much to offset the track

  //getting hit
  float hitX=15;    //offset from regular x
  float hitY=10;   //offset from regular y
  float hitW=60;
  float hitH=140;
  int hitTime=25;  //how many frames to show red upon being hit
  int hitTimer=0;

  //showing the image
  PImage pic;

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=4;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=10;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  //whirlpools
  WhirlPool whirlpools[]=new WhirlPool[6];
  float whirlpoolDamage=2;
  float damage=3;  //damage from touching the beast

  //timing
  int raiseTimes[];
  int nextRaiseTime=0;
  int lowerTimes[];
  int nextLowerTime=0;

  int whirlTimes[];
  int whirlStrength[];
  int nextWhirlTime=0;

  BossWhirlPool() {
    x=xCenter;
    y=yCenter; 

    pic=loadImage("data/pics/whirlpoolMonster.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      int col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    //set up the whirlPools
    float angleSpacing=TWO_PI/whirlpools.length;
    for (int i=0; i<whirlpools.length; i++) {
      whirlpools[i]=new WhirlPool(i*angleSpacing);
    }

    //set timing
    setTiming();
  }

  public void update(int time) {
    //keep it moving up and down
    yCenter=height/2+sin(millis()*ySpeed)*yPadding;
    xCenter=width/2+sin(millis()*xSpeed)*xPadding +xShift;
    //move it around that center point
    x=xCenter+noise((float)millis()*noiseSpeedX, 1000)*xRange*2-xRange;
    y=yCenter+noise((float)millis()*noiseSpeedY, 1000)*yRange*2-yRange;
    //the 1000 just serves to seperate if from the noise being generated for the other monsters

    //if this is at all visible, it is active
    active=raisingY>0;

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;

      //set the whirlpool size
      for (int i=0; i<whirlpools.length; i++) {
        whirlpools[i].imageScale=map(raisingY, 0, pic.height-raisingPadding, 0, 1);
      }

      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //adjust the mask for the image
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1f, millis()*0.001f)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }

    //check for timing related events
    if (!gameOver)
      checkTiming(time);

    //lower the hitTimer
    hitTimer--;

    //update the whirlpools
    for (int i=0; i<whirlpools.length; i++)
      whirlpools[i].update(x, y);
  }

  public void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
    //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }

    //time to update the whirlpools?
    if (nextWhirlTime<whirlTimes.length) {
      if (time>whirlTimes[nextWhirlTime]) {
        //set all fo the whilr pools
        setPools(whirlStrength[nextWhirlTime]);
        nextWhirlTime++;
      }
    }
  }

  public void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();

    //draw the whirlpools
    if (active) {
      tint(255, 255, 255);
      for (int i=0; i<whirlpools.length; i++)
        whirlpools[i].draw();
    }

  }

  //called when the boss gets hurt
  public void hit() {
    hitTimer=hitTime;
  }

  //sets the distance on the pools
  //accepts a number between 0 and 1
  public void setPools(float poolDist) {
    for (int i=0; i<whirlpools.length; i++)
      whirlpools[i].setDist(poolDist);
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  public boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }

  //checks to see if the ship hit a whirlpool
  //takes the ship's hit box in the form of the upper left and lower right corners
  public boolean whirlPoolHitCheck(float x1, float x2, float y1, float y2) {
    //if the monster is not out of the water yet, don't let the whirlpools hurt the player
    if (raisingY<pic.height-raisingPadding)  return false;

    for (int i=0; i<whirlpools.length; i++) {
      if (whirlpools[i].checkHit(x1, x2, y1, y2))  return true;
    }
    //nothing was hit, so return false
    return false;
  }

  //causes the monster to sink or rise
  public void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

  //returns true if the monster's hit box is overlapping with the ships
  //box is given by top left corner and bottom right corner
  public boolean isBoxTouching(float x1, float x2, float y1, float y2) {
    //if the monster isn't active, don't register collisions
    if (!active)  return false;

    //need to get the corners of the monsters box
    float myX1=x+hitX-hitW/2;
    float myX2=x+hitX+hitW/2;
    float myY1=y+hitY-hitH/2;
    float myY2=y+hitY+hitH/2;

    if (y2 < myY1) return false;
    if (y1 > myY2) return false;

    if (x2 < myX1) return false;
    if (x1 > myX2) return false;

    return true;
  }

  //hardcoding all of the timing info for the boss in millis
  public void setTiming() {
    raiseTimes=new int[3];
    lowerTimes=new int[3];

    raiseTimes[0]=19120;
    lowerTimes[0]=25500;
    raiseTimes[1]=69500;
    lowerTimes[1]=81500;
    raiseTimes[2]=113900;
    lowerTimes[2]=150000;

    //using array list for now, because it makes it a bit easier to add
    ArrayList<Integer> whirlTimesList=new ArrayList<Integer>();
    ArrayList<Integer> whirlStrengthList=new ArrayList<Integer>();

    //FIRST APPEARANCE

    //set up drum hits
    int drumStartTime=69900;
    int drumEndTime=80000;
    int drumCycleTime=1565;  //how long it takes to go through a fulls et of drum hits
    //timing within the cycle
    int cycleTiming[]=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=400;
    cycleTiming[2]=566;
    cycleTiming[3]=800;
    cycleTiming[4]=1220;
    //power for each hit
    int cyclePower[]=new int[5];
    cyclePower[0]=150;
    cyclePower[1]=20;
    cyclePower[2]=20;
    cyclePower[3]=80;
    cyclePower[4]=80;

    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<cycleTiming.length; c++) {
        whirlTimesList.add(i+cycleTiming[c]);
        whirlStrengthList.add(cyclePower[c]);
      }
    }
    //one hit at the end of the segment
    whirlTimesList.add(81000);
    whirlStrengthList.add(200);


    //SECOND APPEARANCE
    drumStartTime=114080;  //115650;
    drumEndTime=137300;  //when the heavier drums kick in
    drumCycleTime=1575;  //how long it takes to go through a fulls et of drum hits

    //timing within the cycle
    cycleTiming=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=100;
    cycleTiming[2]=460;
    cycleTiming[3]=860;
    cycleTiming[4]=1260;
    //power for each hit
    cyclePower=new int[5];
    cyclePower[0]=60;
    cyclePower[1]=60;
    cyclePower[2]=40;
    cyclePower[3]=40;
    cyclePower[4]=40;

    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<cycleTiming.length; c++) {
        //need to sip the "Resua get down you'll be blown to simtherens bit" from 2:04.94 to 2:06.28
        if (i+cycleTiming[c]<124940 || i+cycleTiming[c]> 126280){
          whirlTimesList.add(i+cycleTiming[c]);
          whirlStrengthList.add(cyclePower[c]);
        }
      }
    }
    
    //add the hits during the Rensua bit
    int renHits[]=new int[4];
    renHits[0]=125430;
    renHits[1]=126000;
    renHits[2]=126180;
    renHits[3]=126350;
    int renIndex=0;
    //need to find where the other drums ended
    while(whirlTimesList.get(renIndex).intValue()<renHits[0])
      renIndex++;
    //now add them to the list of times starting at that spot
    for (int i=0; i<4; i++){
      whirlTimesList.add(renIndex+i,renHits[i]);
      whirlStrengthList.add(renIndex+i,100);
    }
    
    //drums have a steady kick leading into the heavier section
    int steadyStart=137580;
    int steadyEnd  =139010;
    int steadyPeriod  =210; 
    for (int i=steadyStart; i<steadyEnd; i+=steadyPeriod) {
      whirlTimesList.add(i);
      whirlStrengthList.add(70);
    }
    
    //heavier section
    drumStartTime=139400;  
    drumEndTime=149500;  //when the heavier drums kick in
    drumCycleTime=1579;  //how long it takes to go through a fulls et of drum hits

    //timing within the cycle MAYBE WE ONLY NEED 3, NOT ALL 5
    cycleTiming=new int[5];
    cycleTiming[0]=0;
    cycleTiming[1]=372;
    cycleTiming[2]=690;
    cycleTiming[3]=998;
    cycleTiming[4]=1684;
    //power for each hit
    cyclePower=new int[5];
    cyclePower[0]=40;
    cyclePower[1]=40;
    cyclePower[2]=40;
    cyclePower[3]=40;
    cyclePower[4]=40;
    
    for (int i=drumStartTime; i<drumEndTime; i+=drumCycleTime) {
      //add a whirlpool hit for each drum hit in the cycle
      for (int c=0; c<3; c++) {
        whirlTimesList.add(i+cycleTiming[c]);
        whirlStrengthList.add(cyclePower[c]);
      }
    }
    
    

    //convert the arraylists to good old arrays after
    whirlTimes=new int[whirlTimesList.size()];
    whirlStrength=new int[whirlStrengthList.size()];
    for (int i=0; i<whirlTimesList.size(); i++) {
      whirlTimes[i]= whirlTimesList.get(i).intValue();
      whirlStrength[i]= whirlStrengthList.get(i).intValue();
    }
  }

  //advances all timers to this point in the song
  public void skipTo(int time) {
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;

    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;

    for (int i=0; i<whirlTimes.length;i++)
      if (time > whirlTimes[i])  nextWhirlTime=i;
  }
}

class Bullet{
  float x,y;  //location
  
  //how far from the center of the ship to start it
  float inFrontOffset=35;  
  float inBackOffset=25;
  
  float speed=8;  //how fast it moves
  PImage pic;   //image of the cannon ball
  
  //hitting things
  float hitSize=14;
  float damage=1;
  
  //rotation for style
  float curAngle=0;
  float rotSpeed=0.2f;  //how fast to spin
  
  boolean killMe;  //flag becomes true if the cannon ball is off screen
  
  boolean inFront;  //is the bullet in front of or behind the ship?
 
  Bullet(float _x, float _y, boolean _inFront){
    x=_x;
    inFront=_inFront;
    if (inFront)  y=_y+inFrontOffset;
    else          y=_y+inBackOffset;
    
    killMe=false;
    
    //load the image
    pic=loadImage("data/pics/cannonball.png");
  } 
  
  public void update(){
    //move the ball
    x+=speed;
    //check if it is offscreen
    if (x>width+100)  killMe=true;
    
    //rotate
    curAngle+=rotSpeed;
  }
  
  public void draw(){
    imageMode(CENTER); 
    pushMatrix();
    translate(x,y);
    rotate(curAngle);
    tint(255,255,255);
    image(pic,0,0);
    popMatrix();
    
    //draw the hit box
//     noFill();
//    stroke(255, 0, 0);
//    ellipse(x,y,hitSize,hitSize);
  }
  
  
}
//this is the giant head the kills the player

class Finisher {
  float x, y;  //location
  boolean active;  //keeps track of if this thing is currently in play
  float yOffset=-30;  //needs to be raised a bit form ship's y pos
  
  //showing the image
  PImage pic;

  float killHeight=60;  //what the raisingY needs to be to kill the ship
  boolean killShip;  //becomes true when the head is out of the water enough to kill the ship

  //raising out of the water
  float raisingDir=0;  //-1 means raising up & 1 means going down
  float raisingSpeed=10;  //how fast it rises or sinks
  float raisingY=0;      //how far out of the water the boss is
  float raisingPadding=50;  //how much of the image of to leave in the water
  float waveHeight=10;    //how high the water should come up
  //image mask
  int maskValues[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored

  //timing
  int nextRaiseTime=0;
  int raiseTimes[];
  int lowerTimes[];
  int nextLowerTime=0;

  Finisher() { 
    killShip=false;
    pic=loadImage("data/pics/whirlpoolMonsterFinisher.png");
    maskValues=new int[pic.width*pic.height];

    //go thorugh the image and figure out which pixels must remain transparent (this is because masking in Processing is dumb as hell)
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      int col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }

    setTimes();
  }

  public void update(int time) {

    checkTiming(time);

   //keep the finisher on the player
   x=ship.x;
   y=ship.y+yOffset;


    //if this is at all visible, it is active
    active=raisingY>0;

    //if it is raising or lowering, adjust the mask
    if (raisingDir!=0) {
      raisingY+= raisingSpeed*raisingDir;

      //check if we reached the top or bottom of the image
      if (raisingY>=pic.height-raisingPadding || raisingY<=0)  raisingDir=0;
    }

    //adjust the mask for the image
    for (int c=0; c<pic.width; c++) {
      //figure out where in this column the water is in order to mask the things beneath it
      //add some noise to the value to simulate waves
      int yCutOff=(int) (raisingY-noise(c*0.1f, millis()*0.001f)*waveHeight );


      for (int r=0; r<pic.height; r++) {
        int pos=r*pic.width+c;  //where in the array it is
        //if this is in the transparent part of the image, keep it that way
        if (maskTrans[pos])
          maskValues[pos]=0;
        else {
          //otherwise, adjust it depending on the raisingY
          if (r<=yCutOff)  maskValues[pos]=255;
          else             maskValues[pos]=0;
        }
      }
    }
    
    
    //is it time to kill the ship?
    if (raisingY>killHeight)  killShip=true;

  }

  public void checkTiming(int time) {
    //time to raise?
    if (nextRaiseTime<raiseTimes.length) {
      if (time>raiseTimes[nextRaiseTime]) {
        raise(1);
        nextRaiseTime++;
      }
    }
     //time to lower?
    if (nextLowerTime<lowerTimes.length) {
      if (time>lowerTimes[nextLowerTime]) {
        raise(-1);
        nextLowerTime++;
      }
    }
  }

  public void draw() {

    imageMode(CENTER); 
    pushMatrix();
    translate(x, y+pic.height-raisingY);
    tint(255, 255, 255);
    pic.mask(maskValues);
    image(pic, 0, 0);
    popMatrix();
  }

  //causes the monster to sink or rise
  public void raise(int dir) {
    raisingDir=dir;
    //make sure it is in place
    raisingY=constrain(raisingY, 0, pic.height-raisingPadding);
  }

   //hard code the times that this thing needs to charge
  public void setTimes() {
    raiseTimes=new int[1];
    raiseTimes[0]=167820;
    
    lowerTimes=new int[1];
    lowerTimes[0]=169050;
    
  }
  
  //advances all timers to this point in the song
  public void skipTo(int time) {
    for (int i=0; i<raiseTimes.length; i++)
      if (time > raiseTimes[i])  nextRaiseTime=i;
    for (int i=0; i<lowerTimes.length; i++)
      if (time > lowerTimes[i])  nextLowerTime=i;
  }
  
}

class Ship {
  float x, y;  //lcoation

  //movment
  float moveSpeed=5;  //speed (no surprise there)
  float padding=50;    //how close to the edges to let the ship get

  //bobbing
  float bobAngle=0;
  float bobSpeed=0.005f;
  float bobMax=0.1f;  //angle in radians that the boat can bob

  //capping the fire rate
  int shotTime=20;  //how many frames between shots
  int shotTimer=0;  //the timer

  //getting hit
  float hitX=0;    //offset from regular x
  float hitY=30;   //offset from regular y
  float hitW=80;
  float hitH=30;
  int hitTime=30;  //how many frames to show red upon being hit
  int hitTimer=0;

  //graphics
  PImage pic; //pic
  
  //is the ship dead?
  boolean isDead;
  int sinkingMask[];
  boolean maskTrans[];      //the pixels that need to remain transparent
  int transThreshold=100;   //transparency bellow this will be ignored
  float sinkingY;      //this should start off being the height of the ship
  float sinkingSpeed=4;

  Ship() {
    x=100;
    y=height/2; 

    isDead=false;

    //load in the pic
    pic=loadImage("data/pics/ship.png");
    sinkingMask=new int[pic.width*pic.height];
    sinkingY=pic.height;
    
    //set which pixels can be transparent
    pic.loadPixels();  //get the pixels from the image
    maskTrans=new boolean[pic.width*pic.height];
    for (int i=0; i<maskTrans.length; i++) {
      maskTrans[i]=true;
      int col=pic.pixels[i];
      int a= (col >> 24) & 0xFF;  //bit shift for the alpha value
      if (a>transThreshold) {
        maskTrans[i]=false;
      }
    }
    
  }

  public void update() {
    shotTimer++;  //increase the shot timer
    hitTimer--;  //reduce the hit timer

    //keep the ship in bounds
    if (x<padding)        x=padding;
    if (x>width-padding)  x=width-padding;
    if (y<padding)        y=padding;
    if (y>height-padding) y=height-padding;

    //have the boat bob a bit
    bobAngle+=bobSpeed;
    if (bobAngle>bobMax || bobAngle<-bobMax)
      bobSpeed*=-1;
      
    //if it is dead, sink the ship using the mask
    if (isDead){
      sinkingY-=sinkingSpeed;
      for (int c=0; c<pic.width; c++){
        for (int r=0; r<pic.height; r++){
          int pos=r*pic.width+c;
          
          //assume we want to draw this pixel
          sinkingMask[pos]=255;
          
          //check if this is one of the values that must be transparent
          if (maskTrans[pos])  sinkingMask[pos]=0;
          
          //if this is below the sinking level, make it transparent
          if (r>sinkingY)  sinkingMask[pos]=0;
        }
      }
    }
  }

  public void draw() {
    imageMode(CENTER); 

    //rotate a bit for bob

    pushMatrix();
    translate(x, y);
    //if the player is dying, move the ship down
    if (isDead) translate(0,pic.height-sinkingY);
    
    //mask some of the image if it is sinking
    if (isDead)  pic.mask(sinkingMask);

    //translate(0,pic.width/2);
    rotate(bobAngle);
    //translate(0,-pic.width/2);
    if (hitTimer>0 && frameCount%10>5)  tint(255, 0, 0);
    else             tint(255, 255, 255);
    image(pic, 0, 0);
    popMatrix();

    //draw hitRange
//    rectMode(CENTER);
//    noFill();
//    stroke(0, 255, 0);
//    rect(x+hitX, y+hitY, hitW, hitH);
  }

  //moves the ship
  public void moveShip(int horzDir, int vertDir) {
    if (!isDead){
      x+=moveSpeed*horzDir;
      y+=moveSpeed*vertDir;
    }
  }

  //handles shooting a cannonball
  //actual bullet spawn is handled in main, this handles other effects
  public void shoot() {
    shotTimer=0;
  }

  //called when the ship gets hurt
  public void hit() {
    if (!isDead){
      hitTimer=hitTime;
      if (playerHealth<=0)
        isDead=true;
    }
      
  }

  //returns true if the given location is within the hit box
  //assumes circles are being checked against the box
  public boolean hitCheck(float otherX, float otherY, float range) {
    return (otherX+range>x+hitX-hitW/2 && otherX-range<x+hitX+hitW/2 && otherY+range>y+hitY-hitH/2 && otherY-range<y+hitY+hitH/2);
  }
}

//whirl pools swirl around one of the monsters
class WhirlPool {
  float x, y;  //location
  float hitSize;

  float retreatSpeed=6;  //how quickly it moves back to the monster
  float minDist=150;     //how close it can get to the monster
  float distFromCenter=minDist;  //reset by the monster

  //rotating around the center
  float curAngle;
  float rotSpeed=-0.01f;
  
  //spinning the pic
  float spinAngle=0;
  float spinSpeed=0.1f;
  
  //the pic grows and shrinks when the monster is rising or lowering
  float imageScale=1;

  PImage pic;

  float crunch=0.5f;  //how much to shrink the image verticlly 
  
  WhirlPool(float startAngle){
     pic=loadImage("data/pics/whirlpool.png");

    //hit range will be half the size of the image
    hitSize=pic.width/4;
    
    curAngle=startAngle;
  }

  public void update(float centerX, float centerY) {
    //keep rotating around the monster
    curAngle+=rotSpeed;
    //move the whirlpool around the monster's center
    x=centerX+cos(curAngle)*distFromCenter;
    y=centerY+sin(curAngle)*distFromCenter;
    
    //keep the thing spinning
    spinAngle+=spinSpeed;
    
    //have the pool retreat towards the min dist
    if (distFromCenter>minDist)  distFromCenter-=retreatSpeed;
  }

  public void draw() {
    pushMatrix();
    translate(x,y);
    scale(1,crunch);
    rotate(spinAngle);
    scale(imageScale,imageScale);
    image(pic,0,0);
    popMatrix();
    
    //draw the hit box
//    noFill();
//    stroke(255, 0, 0);
//    ellipse(x,y,hitSize,hitSize);
  }

  public void setDist(float newDist) {
    distFromCenter+=newDist;
    distFromCenter=max(minDist,distFromCenter);
  }
  
  //checks to see if the whirlpool is touching the ship's hit box
  //takes the ship's hit box in the form of the upper left and lower right corners
  public boolean checkHit(float x1, float x2, float y1, float y2){
    return (x+hitSize>x1 && x-hitSize<x2 && y+hitSize>y1 && y-hitSize<y2);
  }
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#FFFFFF", "SonsOfTheWaves" });
  }
}
